import random
import string
import pandas as pd


class Coverage:
    def __init__(
        self,
        dataframe,
        demand_col=None,
        demand_name="demand",
        supply_name=None,
        coverage_type="binary",
    ):
        """
        An object that stores the relationship between a set of demand locations and a set of supply locations.
        Use this initializer if the coverage matrix has already been created, otherwise this can be created from two
        geodataframes using the :meth:`~allagash.coverage.Coverage.from_geodataframes` or
        :meth:`~allagash.coverage.Coverage.from_spatially_enabled_dataframes` factory methods.

        .. code-block:: python

            Coverage.from_geodataframes(df1, df2, "Demand_Id", "Supply_Id")

        :param ~pandas.DataFrame dataframe: A dataframe containing a matrix of demand (rows) and supply (columns).
                                        An additional column containing the demand values can optionally be provided.
        :param str demand_col: (optional) The name of the column storing the demand value.
        :param str demand_name: (optional) The name of the demand to use. If not supplied, 'demand' is used'.
        :param str supply_name: (optional) The name of the supply to use. If not supplied, a random name is generated.
        :param str coverage_type: (optional) The type of coverage this represents. If not supplied, the default is
                                  "binary". Options are "binary" and "partial".
        """
        self._validate_init(
            coverage_type, dataframe, demand_col, demand_name, supply_name
        )
        self._demand_col = demand_col
        self._dataframe = dataframe
        if not demand_name:
            self._demand_name = "".join(random.choices(string.ascii_uppercase, k=6))
        else:
            self._demand_name = demand_name
        if not supply_name:
            self._supply_name = "".join(random.choices(string.ascii_uppercase, k=6))
        else:
            self._supply_name = supply_name
        self._coverage_type = coverage_type.lower()

    @staticmethod
    def _validate_init(coverage_type, dataframe, demand_col, demand_name, supply_name):
        if not isinstance(dataframe, pd.DataFrame):
            raise TypeError(
                f"Expected 'Dataframe' type for dataframe, got '{type(dataframe)}'"
            )
        if not isinstance(demand_col, str) and demand_col is not None:
            raise TypeError(
                f"Expected 'str' type for demand_col, got '{type(demand_col)}'"
            )
        if not isinstance(demand_name, str) and demand_name is not None:
            raise TypeError(
                f"Expected 'str' type for demand_name, got '{type(demand_name)}'"
            )
        if not isinstance(supply_name, str) and supply_name is not None:
            raise TypeError(
                f"Expected 'str' type for supply_name, got '{type(supply_name)}'"
            )
        if not isinstance(coverage_type, str):
            raise TypeError(
                f"Expected 'str' type for coverage_type, got '{type(coverage_type)}'"
            )
        if demand_col and demand_col not in dataframe.columns:
            raise ValueError(f"'{demand_col}' not in dataframe")
        if coverage_type.lower() not in ("binary", "partial"):
            raise ValueError(f"Invalid coverage type '{coverage_type}'")
        if coverage_type.lower() == "partial" and demand_col is None:
            raise ValueError(
                "'demand_col' is required when generating partial coverage"
            )

    @property
    def df(self):
        """

        :return: The geodataframe the dataset is based on
        :rtype: ~geopandas.GeoDataFrame
        """
        return self._dataframe

    @property
    def demand_name(self):
        """

        :return: The name of the demand
        :rtype: str
        """
        return self._demand_name

    @property
    def supply_name(self):
        """

        :return: The name of the supply
        :rtype: str
        """
        return self._supply_name

    @property
    def coverage_type(self):
        """

        :return: The type of coverage
        :rtype: str
        """
        return self._coverage_type

    @property
    def demand_col(self):
        """

        :return: The name of the demand column in the underlying dataframe
        :rtype: str or None
        """
        return self._demand_col

    @classmethod
    def from_geodataframes(
        cls,
        demand_df,
        supply_df,
        demand_id_col,
        supply_id_col,
        demand_name="demand",
        supply_name=None,
        demand_col=None,
        coverage_type="binary",
    ):
        """
        Creates a new Coverage from two GeoDataFrames representing the demand and supply locations. The coverage
        is determined by intersecting the two dataframes.

        :param ~geopandas.GeoDataFrame demand_df: The GeoDataFrame containing the demand locations
        :param ~geopandas.GeoDataFrame supply_df: The GeoDataFrame containing the supply locations
        :param str demand_id_col: The name of the column that has unique identifiers for the demand locations
        :param str supply_id_col: The name of the column that has unique identifiers for the supply locations
        :param str demand_name: (optional) The name of the demand to use. If not supplied, 'demand' is used.
        :param str supply_name: (optional) The name of the supply to use. If not supplied, a random name is generated.
        :param str demand_col: (optional) The name of the column that stores the amount of demand for the demand
                                          locations. Required if generating partial coverage.
        :param str coverage_type: (optional) The type of coverage this represents. If not supplied, the default is
                                  "binary". Options are "binary" and "partial".

        :return: The coverage
        :rtype: ~allagash.coverage.Coverage
        """
        cls._validate_from_geodataframes(
            coverage_type,
            demand_col,
            demand_df,
            demand_id_col,
            demand_name,
            supply_df,
            supply_id_col,
        )

        data = []
        if coverage_type.lower() == "binary":
            for index, row in demand_df.iterrows():
                contains = supply_df.geometry.contains(row.geometry).tolist()
                if demand_col:
                    contains.insert(0, row[demand_col])
                # Add the id column to the end, it will be used as index and removed later
                contains.append(row[demand_id_col])
                data.append(contains)
        elif coverage_type.lower() == "partial":
            for index, row in demand_df.iterrows():
                demand_area = row.geometry.area
                intersection_area = supply_df.geometry.intersection(
                    row.geometry
                ).geometry.area
                partial_coverage = (
                    (intersection_area / demand_area) * row[demand_col]
                ).tolist()
                if demand_col:
                    partial_coverage.insert(0, row[demand_col])
                partial_coverage.insert(0, row[demand_id_col])
                data.append(partial_coverage)
        else:
            raise ValueError(f"Invalid coverage type '{coverage_type}'")
        columns = supply_df[supply_id_col].tolist()
        if demand_col:
            columns.insert(0, demand_col)
        # id column will be used as index when dataframe is created
        columns.append(demand_id_col)
        # Set index after to avoid issue with multiindex being created
        df = pd.DataFrame.from_records(data, columns=columns).set_index(demand_id_col)
        return Coverage(
            df,
            demand_col=demand_col,
            demand_name=demand_name,
            supply_name=supply_name,
            coverage_type=coverage_type,
        )

    @classmethod
    def from_spatially_enabled_dataframes(
        cls,
        demand_df,
        supply_df,
        demand_id_col,
        supply_id_col,
        demand_name="demand",
        supply_name=None,
        demand_col=None,
        coverage_type="binary",
        demand_geometry_col="SHAPE",
        supply_geometry_col="SHAPE",
    ):
        """
        Creates a new Coverage from two spatially enabled (arcgis) dataframes representing the demand and supply locations.
        The coverage is determined by intersecting the two dataframes.

        :param ~pandas.DataFrame demand_df: The spatially enabled dataframe containing the demand locations
        :param ~pandas.DataFrame supply_df: The spatially enavled dataframe containing the supply locations
        :param str demand_id_col: The name of the column that has unique identifiers for the demand locations
        :param str supply_id_col: The name of the column that has unique identifiers for the supply locations
        :param str demand_name: (optional) The name of the demand to use. If not supplied, 'demand' is used'.
        :param str supply_name: (optional) The name of the supply to use. If not supplied, a random name is generated.
        :param str demand_col: (optional) The name of the column that stores the amount of demand for the demand
                                          locations. Required if generating partial coverage.
        :param str coverage_type: (optional) The type of coverage this represents. If not supplied, the default is
                                  "binary". Options are "binary" and "partial".
        :param str demand_geometry_col: (optional) The name of the field storing the geometry in the demand dataframe.
                                        If not supplied, the default is "SHAPE".
        :param str supply_geometry_col: (optional) The name of the field storing the geometry in the supply dataframe.
                                        If not supplied, the default is "SHAPE".
        :return: The coverage
        :rtype: ~allagash.coverage.Coverage
        """

        cls._validate_from_spatially_enabled_dataframes(
            coverage_type,
            demand_col,
            demand_df,
            demand_id_col,
            demand_name,
            supply_df,
            supply_id_col,
            demand_geometry_col,
            supply_geometry_col,
        )
        data = []
        if coverage_type.lower() == "binary":
            for index, row in demand_df.iterrows():
                contains = (
                    supply_df[supply_geometry_col]
                    .geom.contains(row[demand_geometry_col])
                    .tolist()
                )
                if demand_col:
                    contains.insert(0, row[demand_col])
                # Add the id column to the end, it will be used as index and removed later
                contains.append(row[demand_id_col])
                data.append(contains)
        elif coverage_type.lower() == "partial":
            for index, row in demand_df.iterrows():
                partial_coverage = []
                demand_area = row[demand_geometry_col].area
                # Cannot vectorize this because if the intersection returns an empty polygon with rings
                # The conversion to shapely fails when trying to get the area
                for _, s_row in supply_df.iterrows():
                    intersection = s_row[supply_geometry_col].intersect(
                        row[demand_geometry_col]
                    )
                    area = intersection.area if not intersection.is_empty else 0
                    partial_coverage.append((area / demand_area) * row[demand_col])
                if demand_col:
                    partial_coverage.insert(0, row[demand_col])
                # Add the id column to the end, it will be used as index and removed later
                partial_coverage.append(row[demand_id_col])
                data.append(partial_coverage)
        else:
            raise ValueError(f"Invalid coverage type '{coverage_type}'")
        columns = supply_df[supply_id_col].tolist()
        if demand_col:
            columns.insert(0, demand_col)
        columns.append(demand_id_col)
        df = pd.DataFrame.from_records(data, columns=columns).set_index(demand_id_col)
        return Coverage(
            df,
            demand_col=demand_col,
            demand_name=demand_name,
            supply_name=supply_name,
            coverage_type=coverage_type,
        )

    @classmethod
    def _validate_from_geodataframes(
        cls,
        coverage_type,
        demand_col,
        demand_df,
        demand_id_col,
        demand_name,
        supply_df,
        supply_id_col,
    ):
        if not isinstance(demand_df, pd.DataFrame):
            raise TypeError(
                f"Expected 'Dataframe' type for demand_df, got '{type(demand_df)}'"
            )
        if not isinstance(supply_df, pd.DataFrame):
            raise TypeError(
                f"Expected 'Dataframe' type for supply_df, got '{type(supply_df)}'"
            )
        if not isinstance(demand_id_col, str):
            raise TypeError(
                f"Expected 'str' type for demand_id_col, got '{type(demand_id_col)}'"
            )
        if not isinstance(supply_id_col, str):
            raise TypeError(
                f"Expected 'str' type for demand_id_col, got '{type(supply_id_col)}'"
            )
        if not isinstance(demand_name, str) and demand_name is not None:
            raise TypeError(
                f"Expected 'str' type for demand_name, got '{type(demand_name)}'"
            )
        if not isinstance(coverage_type, str):
            raise TypeError(
                f"Expected 'str' type for coverage_type, got '{type(coverage_type)}'"
            )
        if demand_col and demand_col not in demand_df.columns:
            raise ValueError(f"'{demand_col}' not in dataframe")
        if demand_id_col and demand_id_col not in demand_df.columns:
            raise ValueError(f"'{demand_id_col}' not in dataframe")
        if supply_id_col and supply_id_col not in supply_df.columns:
            raise ValueError(f"'{supply_id_col}' not in dataframe")
        if coverage_type.lower() not in ("binary", "partial"):
            raise ValueError(f"Invalid coverage type '{coverage_type}'")
        if coverage_type.lower() == "partial" and demand_col is None:
            raise ValueError("demand_col is required when generating partial coverage")

    @classmethod
    def _validate_from_spatially_enabled_dataframes(
        cls,
        coverage_type,
        demand_col,
        demand_df,
        demand_id_col,
        demand_name,
        supply_df,
        supply_id_col,
        demand_geometry_col,
        supply_geometry_col,
    ):
        if not isinstance(demand_df, pd.DataFrame):
            raise TypeError(
                f"Expected 'Dataframe' type for demand_df, got '{type(demand_df)}'"
            )
        if not isinstance(supply_df, pd.DataFrame):
            raise TypeError(
                f"Expected 'Dataframe' type for supply_df, got '{type(supply_df)}'"
            )
        if not isinstance(demand_id_col, str):
            raise TypeError(
                f"Expected 'str' type for demand_id_col, got '{type(demand_id_col)}'"
            )
        if not isinstance(supply_id_col, str):
            raise TypeError(
                f"Expected 'str' type for demand_id_col, got '{type(supply_id_col)}'"
            )
        if not isinstance(demand_name, str) and demand_name is not None:
            raise TypeError(
                f"Expected 'str' type for demand_name, got '{type(demand_name)}'"
            )
        if not isinstance(coverage_type, str):
            raise TypeError(
                f"Expected 'str' type for coverage_type, got '{type(coverage_type)}'"
            )
        if demand_col and demand_col not in demand_df.columns:
            raise ValueError(f"'{demand_col}' not in dataframe")
        if demand_id_col and demand_id_col not in demand_df.columns:
            raise ValueError(f"'{demand_id_col}' not in dataframe")
        if supply_id_col and supply_id_col not in supply_df.columns:
            raise ValueError(f"'{supply_id_col}' not in dataframe")
        if demand_geometry_col and demand_geometry_col not in demand_df.columns:
            raise ValueError(f"'{demand_geometry_col}' not in dataframe")
        if supply_geometry_col and supply_geometry_col not in supply_df.columns:
            raise ValueError(f"'{supply_geometry_col}' not in dataframe")
        if coverage_type.lower() not in ("binary", "partial"):
            raise ValueError(f"Invalid coverage type '{coverage_type}'")
        if coverage_type.lower() == "partial" and demand_col is None:
            raise ValueError("demand_col is required when generating partial coverage")



import operator
import pulp
from .coverage import Coverage
from pandas.api.types import is_string_dtype


class Problem:
    _problem_types = ["lscp", "mclp", "bclp"]
    _delineator = "$"

    def __init__(self, pulp_problem, coverages, problem_type):
        """
        A representation of the linear programming problem that can be solved.
        This is not intended to be created on it's own but rather from one of the factory methods
        :meth:`~allagash.problem.Problem.lscp` or :meth:`~allagash.problem.Problem.mclp`

        .. code-block:: python

            Problem.lscp(coverage)
            Problem.lscp([coverage1, coverage2])

        :param ~pulp.LpProblem pulp_problem: The pulp problem that will be solved
        :param list[~allagash.coverage.Coverage] coverages: The coverages that were used to build the problem
        :param str problem_type: The type of problem that was generated
        """
        self._validate(pulp_problem, coverages, problem_type)
        self._pulp_problem = pulp_problem
        if isinstance(coverages, Coverage):
            self._coverages = [coverages]
        else:
            self._coverages = coverages
        self._problem_type = problem_type.lower()

    def _validate(self, problem, coverages, problem_type):
        if not isinstance(problem, pulp.LpProblem):
            raise TypeError(
                f"Expected 'LpProblem' type for problem, got '{type(problem)}'"
            )
        if not isinstance(problem_type, str):
            raise TypeError(
                f"Expected 'str' type for problem_type, got '{type(problem_type)}'"
            )
        if problem_type.lower() not in self._problem_types:
            raise ValueError(f"Invalid problem_type: '{problem_type}'")
        if not isinstance(coverages, (list, Coverage)):
            raise TypeError(
                f"Expected 'Coverage' or 'list' type for coverages, got '{type(coverages)}'"
            )

    @property
    def pulp_problem(self):
        """

        :return: The pulp problem
        :rtype: ~pulp.LpProblem
        """
        return self._pulp_problem

    @property
    def coverages(self):
        """

        :return: The coverage used to create the problem
        :rtype: list[~allagash.coverage.Coverage]
        """
        return self._coverages

    @property
    def problem_type(self):
        """

        :return: The type of problem that this is
        :rtype: str
        """
        return self._problem_type

    def solve(self, solver):
        """

        :param ~pulp.solvers.LpSolver solver: The solver to use for this problem
        :return: The solution for this problem
        :rtype: ~allagash.problem.Problem
        """
        if not isinstance(solver, pulp.LpSolver):
            raise TypeError(
                f"Expected 'LpSolver' type for solver, got '{type(solver)}'"
            )
        self._pulp_problem.solve(solver)
        if self._pulp_problem.status == 0:
            raise NotSolvedException("Unable to solve the problem")
        elif self._pulp_problem.status == -1:
            raise InfeasibleException("Infeasible problem")
        elif self._pulp_problem.status == -2:
            raise UnboundedException("Unbounded problem")
        elif self._pulp_problem.status == -3:
            raise UndefinedException("Undefined problem")
        return self

    @classmethod
    def lscp(cls, coverages):
        """
        Creates a new :class:`~allagash.problem.Problem` object representing the Location Covering Set Problem

        :param list[~allagash.coverage.Coverage] coverages: The coverages to be used to create the problem
        :return: The created problem
        :rtype: ~allagash.problem.Problem
        """
        if not isinstance(coverages, (Coverage, list)):
            raise TypeError(
                f"Expected 'Coverage' or 'list' type for coverages, got '{type(coverages)}'"
            )
        if isinstance(coverages, Coverage):
            coverages = [coverages]
        if not all([c.coverage_type == coverages[0].coverage_type for c in coverages]):
            raise ValueError(
                "Invalid coverages. Coverages must have the same coverage type."
            )
        if coverages[0].coverage_type != "binary":
            raise ValueError("LSCP can only be generated from binary coverage.")
        if not all(x.demand_name == coverages[0].demand_name for x in coverages):
            raise ValueError("All Coverages must have the same 'demand_name'")
        prob = cls._generate_lscp_problem(coverages)
        return Problem(prob, coverages, problem_type="lscp")

    @classmethod
    def bclp(cls, coverages, max_supply):
        """
        Creates a new :class:`~allagash.problem.Problem` object representing the Backup Covering Location Problem

        :param list[~allagash.coverage.Coverage] coverages: The coverages to be used to create the problem
        :param dict[~allagash.coverage.Coverage,int] max_supply: The maximum number of supply locations to allow
        :return: The created problem
        :rtype: ~allagash.problem.Problem
        """
        if not isinstance(coverages, (Coverage, list)):
            raise TypeError(
                f"Expected 'Coverage' or 'list' type for coverages, got '{type(coverages)}'"
            )
        if isinstance(coverages, Coverage):
            coverages = [coverages]
        if not all([c.coverage_type == coverages[0].coverage_type for c in coverages]):
            raise ValueError(
                "Invalid coverages. Coverages must have the same coverage type."
            )
        if coverages[0].coverage_type != "binary":
            raise ValueError("BCLP can only be generated from binary coverage.")
        if not isinstance(max_supply, dict):
            raise TypeError(
                f"Expected 'dict' type for max_supply, got '{type(max_supply)}'"
            )
        for k, v in max_supply.items():
            if not isinstance(k, Coverage):
                raise TypeError(
                    f"Expected 'Coverage' type as key in max_supply, got '{type(k)}'"
                )
            if k.demand_col is None:
                raise TypeError("Coverages used in BCLP must have 'demand_col'")
            if not isinstance(v, int):
                raise TypeError(
                    f"Expected 'int' type as value in max_supply, got '{type(v)}'"
                )
        if not all(x.demand_name == coverages[0].demand_name for x in coverages):
            raise ValueError("All Coverages must have the same 'demand_name'")
        prob = cls._generate_bclp_problem(coverages, max_supply)
        return Problem(prob, coverages, problem_type="bclp")

    @classmethod
    def mclp(cls, coverages, max_supply):
        """
        Creates a new :class:`~allagash.problem.Problem` object representing the Maximum Covering Location Problem

        :param list[~allagash.coverage.Coverage] coverages: The coverages to be used to create the problem
        :param dict[~allagash.coverage.Coverage,int] max_supply: The maximum number of supply locations to allow
        :return: The created problem
        :rtype: ~allagash.problem.Problem
        """
        if not isinstance(coverages, (Coverage, list)):
            raise TypeError(
                f"Expected 'Coverage' or 'list' type for coverages, got '{type(coverages)}'"
            )
        if isinstance(coverages, Coverage):
            coverages = [coverages]
        if not all([c.coverage_type == coverages[0].coverage_type for c in coverages]):
            raise ValueError(
                "Invalid coverages. Coverages must have the same coverage type."
            )
        if coverages[0].coverage_type != "binary":
            raise ValueError("MCLP can only be generated from binary coverage.")
        if not isinstance(max_supply, dict):
            raise TypeError(
                f"Expected 'dict' type for max_supply, got '{type(max_supply)}'"
            )
        for k, v in max_supply.items():
            if not isinstance(k, Coverage):
                raise TypeError(
                    f"Expected 'Coverage' type as key in max_supply, got '{type(k)}'"
                )
            if k.demand_col is None:
                raise TypeError("Coverages used in MCLP must have 'demand_col'")
            if not isinstance(v, int):
                raise TypeError(
                    f"Expected 'int' type as value in max_supply, got '{type(v)}'"
                )
        if not all(x.demand_name == coverages[0].demand_name for x in coverages):
            raise ValueError("All Coverages must have the same 'demand_name'")

        prob = cls._generate_mclp_problem(coverages, max_supply)
        return Problem(prob, coverages, problem_type="mclp")

    @staticmethod
    def _generate_lscp_problem(coverages):  # noqa: C901
        demand_vars = {}
        for c in coverages:
            if c.demand_name not in demand_vars:
                demand_vars[c.demand_name] = {}
            for index, _ in c.df.iterrows():
                name = f"{c.demand_name}{Problem._delineator}{index}"
                demand_vars[c.demand_name][index] = pulp.LpVariable(
                    name, 0, 1, pulp.LpInteger
                )

        supply_vars = {}
        for c in coverages:
            if c.demand_col:
                df = c.df.drop(columns=c.demand_col)
            else:
                df = c.df
            if c.supply_name not in supply_vars:
                supply_vars[c.supply_name] = {}
            for s in df.columns.to_list():
                name = f"{c.supply_name}{Problem._delineator}{s}"
                supply_vars[c.supply_name][s] = pulp.LpVariable(
                    name, 0, 1, pulp.LpInteger
                )

        prob = pulp.LpProblem("LSCP", pulp.LpMinimize)
        to_sum = []
        for _, v in supply_vars.items():
            to_sum.append(v)
        prob += pulp.lpSum(to_sum)

        sums = {}
        for c in coverages:
            if c.demand_name not in sums:
                sums[c.demand_name] = {}
            if c.demand_col:
                df = c.df.drop(columns=c.demand_col)
            else:
                df = c.df
            for index, demand in df.iterrows():
                if index not in sums[c.demand_name]:
                    sums[c.demand_name][index] = []
                cov = demand.T
                for i, value in cov.iteritems():
                    if value is True:
                        sums[c.demand_name][index].append(supply_vars[c.supply_name][i])

        for c in coverages:
            for k, v in demand_vars[c.demand_name].items():
                if not to_sum:
                    sums[c.demand_name][v] = [
                        pulp.LpVariable(
                            f"__dummy{Problem._delineator}{v}", 0, 0, pulp.LpInteger
                        )
                    ]

        for demand_name, v in sums.items():
            for i, to_sum in v.items():
                prob += pulp.lpSum(to_sum) >= 1, f"D{demand_name}{i}"
        return prob

    @staticmethod
    def _generate_bclp_problem(coverages, max_supply):  # noqa: C901
        demand_vars = {}
        for c in coverages:
            if c.demand_name not in demand_vars:
                demand_vars[c.demand_name] = {}
            for index, _ in c.df.iterrows():
                name = f"{c.demand_name}{Problem._delineator}{index}"
                demand_vars[c.demand_name][index] = pulp.LpVariable(
                    name, 0, 1, pulp.LpInteger
                )

        supply_vars = {}
        for c in coverages:
            if c.demand_col:
                df = c.df.drop(columns=c.demand_col)
            else:
                df = c.df
            if c.supply_name not in supply_vars:
                supply_vars[c.supply_name] = {}
            for s in df.columns.to_list():
                name = f"{c.supply_name}{Problem._delineator}{s}"
                supply_vars[c.supply_name][s] = pulp.LpVariable(
                    name, 0, 1, pulp.LpInteger
                )

        # add objective
        prob = pulp.LpProblem("BCLP", pulp.LpMaximize)
        demands = {}
        for c in coverages:
            for _, demand_var in demand_vars[c.demand_name].items():
                d = demand_var.name.split(Problem._delineator)[1]
                if d not in demands:
                    if is_string_dtype(coverages[0].df.index.dtype):
                        query = f"{coverages[0].df.index.name} == '{d}'"
                    else:
                        query = f"{coverages[0].df.index.name} == {d}"
                    v = c.df.query(query)[c.demand_col].tolist()[0]
                    demands[d] = v * demand_var
        to_sum = []
        for k, v in demands.items():
            to_sum.append(v)
        prob += pulp.lpSum(to_sum)

        # coverage constraints
        sums = {}
        for c in coverages:
            if c.demand_name not in sums:
                sums[c.demand_name] = {}
            if c.demand_col:
                df = c.df.drop(columns=c.demand_col)
            else:
                df = c.df
            for index, demand in df.iterrows():
                if index not in sums[c.demand_name]:
                    sums[c.demand_name][index] = [
                        -demand_vars[c.demand_name][index],
                        -1,
                    ]
                cov = demand.T
                for i, value in cov.iteritems():
                    if value is True:
                        sums[c.demand_name][index].append(supply_vars[c.supply_name][i])
        for k, v in sums.items():
            for index, to_sum in v.items():
                prob += pulp.lpSum(to_sum) >= 0, f"D{index}"

        # Number of supply locations
        for c in coverages:
            to_sum = []
            for k, v in supply_vars[c.supply_name].items():
                to_sum.append(v)
            prob += (
                pulp.lpSum(to_sum) <= max_supply[c],
                f"Num{Problem._delineator}{c.supply_name}",
            )
        return prob

    @staticmethod
    def _generate_mclp_problem(coverages, max_supply):  # noqa: C901
        demand_vars = {}
        for c in coverages:
            if c.demand_name not in demand_vars:
                demand_vars[c.demand_name] = {}
            for index, _ in c.df.iterrows():
                name = f"{c.demand_name}{Problem._delineator}{index}"
                demand_vars[c.demand_name][index] = pulp.LpVariable(
                    name, 0, 1, pulp.LpInteger
                )

        supply_vars = {}
        for c in coverages:
            if c.demand_col:
                df = c.df.drop(columns=c.demand_col)
            else:
                df = c.df
            if c.supply_name not in supply_vars:
                supply_vars[c.supply_name] = {}
            for s in df.columns.to_list():
                name = f"{c.supply_name}{Problem._delineator}{s}"
                supply_vars[c.supply_name][s] = pulp.LpVariable(
                    name, 0, 1, pulp.LpInteger
                )

        # add objective
        prob = pulp.LpProblem("MCLP", pulp.LpMaximize)
        demands = {}
        for c in coverages:
            for _, demand_var in demand_vars[c.demand_name].items():
                d = demand_var.name.split(Problem._delineator)[1]
                if d not in demands:
                    if is_string_dtype(coverages[0].df.index.dtype):
                        query = f"{coverages[0].df.index.name} == '{d}'"
                    else:
                        query = f"{coverages[0].df.index.name} == {d}"
                    v = c.df.query(query)[c.demand_col].tolist()[0]
                    demands[d] = v * demand_var
        to_sum = []
        for k, v in demands.items():
            to_sum.append(v)
        prob += pulp.lpSum(to_sum)

        # coverage constraints
        sums = {}
        for c in coverages:
            if c.demand_name not in sums:
                sums[c.demand_name] = {}
            if c.demand_col:
                df = c.df.drop(columns=c.demand_col)
            else:
                df = c.df
            for index, demand in df.iterrows():
                if index not in sums[c.demand_name]:
                    sums[c.demand_name][index] = [-demand_vars[c.demand_name][index]]
                cov = demand.T
                for i, value in cov.iteritems():
                    if value is True:
                        sums[c.demand_name][index].append(supply_vars[c.supply_name][i])
        for k, v in sums.items():
            for index, to_sum in v.items():
                prob += pulp.lpSum(to_sum) >= 0, f"D{index}"

        # Number of supply locations
        for c in coverages:
            to_sum = []
            for k, v in supply_vars[c.supply_name].items():
                to_sum.append(v)
            prob += (
                pulp.lpSum(to_sum) <= max_supply[c],
                f"Num{Problem._delineator}{c.supply_name}",
            )
        return prob

    def selected_supply(self, coverage, operation=operator.eq, value=1):
        """
        Gets the list of the supply locations that were selected when the optimization problem was solved.

        :param ~allagash.coverage.Coverage coverage: The coverage that selected locations may be found in.
        :param function operation: The operation to use when determining whether a location was selected
        :param int value: The value to apply the operation to
        :return: The list of location ids of the selected locations
        :rtype: list
        """
        if self._pulp_problem.status != 1:
            raise RuntimeError("Problem not optimally solved yet")
        from allagash.coverage import Coverage

        if not isinstance(coverage, Coverage):
            raise TypeError(
                f"Expected 'Coverage' type for coverage, got '{type(coverage)}'"
            )
        if not callable(operation):
            raise TypeError(f"Expected callable for operation, got '{type(operation)}'")
        if not isinstance(value, (int, float)):
            raise TypeError(f"Expected 'int' or 'float' for value, got '{type(value)}'")
        ids = []
        for var in self._pulp_problem.variables():
            if var.name.split(self._delineator)[0] == coverage.supply_name:
                if operation(var.varValue, value):
                    ids.append(var.name.split(self._delineator)[1])
        return ids

    def selected_demand(self, coverage):
        """
        Gets the list of the demand locations that were selected when the optimization problem was solved.

        :param ~allagash.coverage.Coverage coverage: The coverage that the demand locations may be found in. If multiple
                coverages were used that have the same demand, locations covered by any other coverages will also
                be returned.
        :return: The list of location ids of the covered locations
        :rtype: list
        """
        if self._pulp_problem.status != 1:
            raise RuntimeError("Problem not optimally solved yet")
        from allagash.coverage import Coverage

        if not isinstance(coverage, Coverage):
            raise TypeError(
                f"Expected 'Coverage' type for coverage, got '{type(coverage)}'"
            )
        if self.problem_type in ["lscp", "bclp"]:
            for c in self.coverages:
                if c.demand_name == c.demand_name:
                    return c.df.index.tolist()
            else:
                raise ValueError(
                    f"Unable to find demand named '{coverage.demand_name}'"
                )
        else:
            ids = []
            for var in self._pulp_problem.variables():
                if var.name.split(self._delineator)[0] == coverage.demand_name:
                    if var.varValue >= 1:
                        ids.append(var.name.split(self._delineator)[1])
            return ids


class NotSolvedException(Exception):
    def __init__(self, message):
        """
        An exception indicating the problem was not solved

        :param str message: A descriptive message about the exception
        """
        super().__init__(message)


class InfeasibleException(Exception):
    def __init__(self, message):
        """
        An exception indicating the problem as an infeasible solution

        :param str message: A descriptive message about the exception
        """
        super().__init__(message)


class UnboundedException(Exception):
    def __init__(self, message):
        """
        An exception indicating the solution is unbounded

        :param str message: A descriptive message about the exception
        """
        super().__init__(message)


class UndefinedException(Exception):
    def __init__(self, message):
        """
        An exception indicating the problem was not solved for an undefined reason

        :param str message: A descriptive message about the exception
        """
        super().__init__(message)



from .problem import (
    Problem,
    UnboundedException,
    UndefinedException,
    InfeasibleException,
    NotSolvedException,
)
from .coverage import Coverage

__all__ = [
    "Problem",
    "UnboundedException",
    "UndefinedException",
    "InfeasibleException",
    "NotSolvedException",
    "Coverage",
]

__version__ = "0.4.3"
